/*
 -- Multicore --
 floaitng point test code:
fp16 format add/sub/mult/divide test for low relative and absolute error
need:


 */

#include "hardware/gpio.h"
#include "hardware/timer.h"

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "stdio.h"
#include <string.h>
#include <pico/multicore.h>
#include <pico/divider.h>
#include "hardware/sync.h"
#include "stdlib.h"
#include "math.h"

// =============================================
// fp16 format
// =============================================
// bit 15 is sign
// 5 bits 14-10 exponent
// 10 bits 9-0 mantissa (with one implied)
// if exp_offset is set to 15, then
// represent one as 0|01111|0000000000
// a zero exponent implies number=0
typedef unsigned short fp16V ;
// Max exp is 31, min is 0
// setting 15 gives a range of +/-1e5 and smallest value ~ 1e-5
#define exp_offset 25
// derive the max/min number bits
#define maxpos 0x7fff
#define maxneg 0xffff
// get the actual max/min for a given exp_offset
float maxnum ;
float minnum ;

//
// =============================================
// type conversions
// =============================================
fp16V IEEE_to_fp16V(float a){
  fp16V b ;
  if(a==0) return 0 ;
  if(fabs(a)>maxnum) a = (a>0)? maxnum : -maxnum ;
  if(fabs(a)<minnum) a = (a>0)? minnum : -minnum ;
  int ia = (*(int*)&a); 
  int sgn = (ia>>31) ;
  int exponent = ((ia>>23) & 0xff) - 127 + exp_offset ; // -127 + 15
  int mantissa = ia & 0x7fffff ;
  // debugging
 // printf("%f %d %x %x\n\r", a, sgn, exponent, mantissa>>13);
  b = (sgn<<15) | ((exponent)<<10)  | (mantissa>>13)  ; // >>23
  return b ;
}
// 
float fp16V_to_IEEE(fp16V a){
  int ia = (int) a ;
  int sgn = (ia>>15) & 0x01 ;
  int exponent = (ia>>10) & 0x1f ;
  if (exponent==0) return 0.0 ;
  int mantissa = ia & 0x3ff ;
  float result = pow(2.0, (float) (exponent - exp_offset));
  result = (1.0+(((float)mantissa) / 1024.0)) * result;
  if(sgn) result = result * (-1);
  return result;
}

// integer conversions
#define int_to_fp16V(a) (IEEE_to_fp16V((float)(a)))
#define fp16V_to_int(a) ((int)fp16V_to_IEEE(a))

// =============================================
// negate and absolute
// =============================================
#define fp16V_neg(a) (((a) & 0x8000)? (a) & 0x7fff : (a) | 0x8000)
#define fp16V_abs(a) ((a) & 0x7fff)

// =============================================
// add/subtract
// =============================================
fp16V fp16V_add(fp16V a, fp16V b){
  // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  int expDiff = aExp - bExp ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
  // debugging
  //printf("%x %x %x %x %x %x \n\r", aFrac, bFrac, aExp, bExp, aSign, bSign);

  // add_sub flag is 1 for subtract
  uint32_t add_sub = aSign ^ bSign ;
  //
  if (aExp==0 | expDiff<-10) return b ;
  if (bExp==0 | expDiff>10) return a ;
  //
  if (expDiff>0){ // a bigger
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      // result is between 1 and 0.0
      uint32_t sum = aFrac - (bFrac>>expDiff) ;
      while (sum < 0x400) {
        aExp--;
        sum <<= 1;
      }
      return (aSign<<15) | (aExp<<10) | (sum & 0x3ff)  ;
    }
    else { //add
      // result is between 1 and 2
      uint32_t sum = aFrac + (bFrac>>expDiff) ;
      if (sum >= 0x800) {
        aExp++;
        sum >>= 1;
      }
      return (aSign<<15) | (aExp<<10) | (sum & 0x3ff)  ;
    }
  }
  //
  else if (expDiff<0){ // b bigger
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      uint32_t sum = bFrac - (aFrac>>-expDiff) ;
      while (sum < 0x400) {
        bExp--;
        sum <<= 1;
      }
      return (bSign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
    else { //add
      uint32_t sum = bFrac + (aFrac>>-expDiff) ;
      if (sum >= 0x800) {
        bExp++;
        sum >>= 1;
      }
      return (bSign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
  }
  else if (expDiff==0){ // 
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      // figure out which mantissa is bigger
      // compute bigger - smaller : result<1
      // normalize
      if (bFrac == aFrac) return 0 ;
      uint32_t sum = (bFrac > aFrac)? bFrac - aFrac : aFrac - bFrac ;
      while (sum < 0x400) {
        bExp--;
        sum <<= 1;
        if (bExp == 0) return 0 ;
      }
      int Sign = (bFrac > aFrac)? bSign : aSign ;
      return (Sign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
    else { //add
      uint32_t sum = bFrac + aFrac ;
      // sum is between 2 and 4
      bExp++;
      sum >>= 1;
      return (bSign<<15) | (bExp<<10) | (sum & 0x3ff)  ;
    }
  }
}

// ==================
fp16V fp16V_sub(fp16V a, fp16V b){
  fp16V diff = fp16V_add(a, fp16V_neg(b)) ;
  return diff ;
}


// =============================================
// multiply
// =============================================
fp16V fp16V_mul(fp16V a, fp16V b){
  // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
  // add exponents an subtract offset
  int expSum = aExp + bExp - exp_offset ;
  // get output sign
  uint32_t Sign = aSign ^ bSign ;
  //
  if (aExp==0 | bExp==0) return 0 ;
  if(expSum > 31) return maxnum ;
  if(aExp + bExp < exp_offset) return 0 ;

  // range is 1 to 4
  uint32_t product = (aFrac * bFrac) >> 10 ;
  if (product >= 0x800) {
    expSum++;
    product >>= 1;
    if(expSum > 31) return maxnum ;
  }
  return (Sign<<15) | (expSum<<10) | (product & 0x3ff) ;
}


// =============================================
// divide
// =============================================
fp16V fp16V_div(fp16V a, fp16V b){
 // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
  // sub exponents an add offset
  int expDiff = aExp - bExp + exp_offset ;
  // get output sign
  uint32_t Sign = aSign ^ bSign ;
  //
  if (aExp==0 | bExp==0) return 0 ;
  if(expDiff< 1) return 0 ;
  if(aExp - bExp > exp_offset) return 0 ;

  // range is 2 to .5
  uint32_t ratio = (aFrac<<10) / bFrac  ;
  if (ratio < 0x400) {
    expDiff--;
    ratio <<= 1;
    if(expDiff < 1) return 0 ;
  }
  return (Sign<<15) | (expDiff<<10) | (ratio & 0x3ff) ;
}


// =============================================
// floating compare
// =============================================
// returns 1 if a >= b otherwise zero
char fp16V_cmp(fp16V a, fp16V b){
 // isolate mantissas and add one
  uint32_t aFrac = (a & 0x3ff) | 0x400 ;
  uint32_t bFrac = (b & 0x3ff) | 0x400 ;
  // isolate the exponents
  int aExp = (a >> 10) & 0x1f ;
  int bExp = (b >> 10) & 0x1f ;
  // isolate the signs
  uint32_t aSign = (a >> 15) & 0x01 ;
  uint32_t bSign = (b >> 15) & 0x01 ;
   // a neg b pos
  if (aSign > bSign) return 0 ; 
  // b neg a pos
  else if (aSign < bSign) return 1 ;
  // both pos
  else if (!aSign) {
    if ((a >= b)) return 1;
    else return 0 ;
  }
  // both neg
  else if (aSign) {
    if ((a <= b)) return 1;
    else return 0 ;
  }   
}


// =============================================
// inverse sqrt
// =============================================
// see https://en.wikipedia.org/wiki/Fast_inverse_square_root
// particularly the section "First approximation of the result"
// The 'magic number' is  1.5 * 2^(num_mantissa_bits) * ((exp_offset)-0.045)
// For fp16; exp offset = 15 ;  num mantissa bits = 10
fp16V fp16V_invsqrt( fp16V number ){
	int i;
	fp16V x2, y;
	const fp16V threehalfs = IEEE_to_fp16V(1.5);
  const fp16V onehalfs = IEEE_to_fp16V(0.5);
  const fp16V small = IEEE_to_fp16V(2*minnum);

  if(!fp16V_cmp(number, small))  return 0 ;
  
	x2 = fp16V_mul(number, onehalfs);
	y  = number;
	i  = * ( int * ) &y;       // evil floating point bit level hacking
                             // exp offset = 15 ;  num mantissa bits = 10
	i  = 22971 - ( i >> 1 );    // 1.5 * 2^(num mantissa bits) * ((exp offset)-0.045)  // what the f?
	y  = (short) i ;           //* ( fp16 * ) &i;
  // now one Newton iteration
	y  = fp16V_mul(y, fp16V_add(threehalfs, fp16V_neg((fp16V_mul(x2,fp16V_mul(y,y))))));   // 1st iteration
//	y  = y * ( threehalfs - ( x2 * y * y ) );   // 2nd iteration, this can be removed
	return y;
}

// =============================================
// floating shift
// =============================================
// positive b means shift left, negaive is shift right
fp16V fp16V_shift( fp16V a, signed char b ){
  // isolate the exponent
  int aExp = (a >> 10) & 0x1f ;
   aExp = (aExp + b) & 0x1f ;
   // mask out exponent and OR  in new one
   a = (a & 0x83ff ) | (aExp << 10) ;
   return a ;
}

// =============================================
// definitions to make arithmetic more compact
// =============================================
#define add fp16V_add
#define sub fp16V_sub
#define cmp fp16V_cmp
#define sft fp16V_shift
#define isqrt fp16V_invsqrt
#define mul fp16V_mul
#define div fp16V_div
#define neg fp16V_neg
#define abs fp16V_abs

// ==========================================
// === protothreads globals
// ==========================================
// protothreads header
#include "pt_cornell_rp2040_v1_3.h"

// ==================================================
// === user's serial input thread
// ==================================================
// test speed of fixed point
static PT_THREAD (protothread_serial(struct pt *pt))
{
  PT_BEGIN(pt);
    static int num_iterations ;
    static int fptime, fp16time ;
    static float a, b, s, f, abs_error, rel_error;
    static fp16V a16, b16, d16;
    static float max_error = 0.01 ;
    static int count ;
    //
    // data structure for interval timer
    PT_INTERVAL_INIT() ;
    //
    /*
    // speed test
      //do fp add
      num_iterations = 1000 ;
    float fp1 = 0;
    fptime = PT_GET_TIME_usec();
    for(int i=0; i<num_iterations; i++){
      fp1 = 1.1 + fp1 ;
    }
    fptime = PT_GET_TIME_usec() - fptime ;
       //do fp216 add 
    d16 = IEEE_to_fp16(0.0);
    fp16 ten = IEEE_to_fp16(1.1);
    fp16time = PT_GET_TIME_usec();
    for(int i=0; i<num_iterations; i++){
      d16 = fp16_add(ten, (d16)) ;
    }
    fp16time = PT_GET_TIME_usec() - fp16time ;

    printf("fptime %d    ieee sum %f \r\n", fptime, fp1) ;
    printf("fp16time %d ffp16 sum %f \r\n", fp16time, fp16_to_IEEE(d16));
    printf("\r\n") ;
    */

    while(1) {
      
        // == test 2 inputs
        maxnum = pow(2, 31-exp_offset+1) ;
        minnum = pow(2, -exp_offset+1) ;
        printf("maxnum=%e minnum=%e", maxnum, minnum) ;
        sprintf(pt_serial_out_buffer, "\n\rinput 2 floats: ");
        serial_write ;
        serial_read ;
        // convert input string to number
        sscanf(pt_serial_in_buffer,"%f %f ", &a, &b) ;
        a16 = IEEE_to_fp16V(a) ; //IEEE_to_fp16(a);
        b16 = IEEE_to_fp16V(b) ; //IEEE_to_fp16(b);
        //a16 = int_to_fp16(a);
        //b16 = int_to_fp16(b);

        sprintf(pt_serial_out_buffer, "ieee to fp16v a16=%e b16=%e \n\r", 
            fp16V_to_IEEE(a16), fp16V_to_IEEE(b16));
        //sprintf(pt_serial_out_buffer, "a16=%d b16=%d \n\r", fp16_to_int(a16), fp16_to_int(b16));
        serial_write ;

        printf( "fp16v to int a16=%d b16=%d \n\r", 
            fp16V_to_int(a16), fp16V_to_int(b16));

        // compare test
        sprintf(pt_serial_out_buffer, "compare16(a>b)=%d \n\r", cmp(a16, b16)) ;
        serial_write ;

        // shift test
        s = fp16V_to_IEEE(sft(a16, (signed char) b)) ;
        sprintf(pt_serial_out_buffer, "shift16(a<<b)=%f \n\r", s) ;
        serial_write ;

        // inv sqrt test
        s = fp16V_to_IEEE(isqrt(a16)) ;
        f = 1/sqrt(a) ;
        sprintf(pt_serial_out_buffer, "invsqrt16(a)=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
        serial_write ;

        // add test
        s = fp16V_to_IEEE(add(a16, b16)) ;
        f = a + b;
        sprintf(pt_serial_out_buffer, "sum16(a+b)=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
        serial_write ;

        // subtract test
        s = fp16V_to_IEEE(sub(a16, b16)) ;
        f = a - b ;
        sprintf(pt_serial_out_buffer, "diff16(a-b)=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
        serial_write ;

        // mult test
        s = fp16V_to_IEEE(mul(a16, b16)) ;
        f = a * b ;
        sprintf(pt_serial_out_buffer, "prod16(a*b)=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
        serial_write ;

        // division test
        s = fp16V_to_IEEE(div(a16, b16)) ;
        f = a / b ;
        sprintf(pt_serial_out_buffer, "ratio16(a/b)=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
        serial_write ;
        
      num_iterations = 100 ;  
      //do fp add
      float fp1 = 0;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = 10.0 + fp1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do fp16 add 
      fp16V d20 = IEEE_to_fp16V(0.0);
      fp16V ten = IEEE_to_fp16V(10.0);
      fp16time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        d20 = add(ten, (d20)) ;
      }
      fp16time = PT_GET_TIME_usec() - fp16time ;

      printf("fptime %d    ieee sum %f \r\n", fptime, fp1) ;
      printf("fp16time %d ffp16 sum %f \r\n", fp16time, fp16V_to_IEEE(d20));
      printf("\r\n") ;

      //do fp mult
      fp1 = 1.0 ;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = 1.1 * fp1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do fp16 mult 
      d20 = IEEE_to_fp16V(1.0);
      fp16V pc = IEEE_to_fp16V(1.1);
      fp16time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        d20 = mul(pc, (d20)) ;
      }
      fp16time = PT_GET_TIME_usec() - fp16time ;

      printf("fptime %d    ieee prod %f \r\n", fptime, fp1) ;
      printf("fp16time %d ffp16 prod %f \r\n", fp16time, fp16V_to_IEEE(d20));
      printf("\r\n") ;

      //do fp div
      fp1 = 1.0 ;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = 1000.0 / fp1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do fp16 div 
      d20 = IEEE_to_fp16V(1.0);
      pc = IEEE_to_fp16V(1000.);
      fp16time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        d20 = div(pc, d20) ;
      }
      fp16time = PT_GET_TIME_usec() - fp16time ;

      printf("fptime %d    ieee ratio %f \r\n", fptime, fp1) ;
      printf("fp16time %d ffp16 ratio %f \r\n", fp16time, fp16V_to_IEEE(d20));
      printf("\r\n") ;

      /*     // brute accuracy test
      for(int i=0; i<10000; i++){
        count++;
        a = ((float)(rand() & 0xfff)/0xfff) * 100.0 - 50.0 ;
        b = ((float)(rand() & 0xfff)/0xfff) * 20.0 - 10.0 ;
        a16 = IEEE_to_fp16(a);
        b16 = IEEE_to_fp16(b);
        //sprintf(pt_serial_out_buffer, "a20=%e b20=%e \n\r", fp16_to_IEEE(a20), fp16_to_IEEE(b20));
        //serial_write ;

        // add accuracy test
        s = fp16_to_IEEE(fp16_add(a16, b16)) ;
        f = a + b;
        abs_error = fabs(f-s) ; 
        rel_error = abs_error/fabs(f) ;
        if ((rel_error > max_error) & (abs_error > max_error)){
            sprintf(pt_serial_out_buffer, "sum16=%f error=%f rel.error=%f\n\r", s, abs_error, rel_error);
            serial_write ;
        }

        // subtract test
        s = fp16_to_IEEE(fp16_sub(a16, b16)) ;
        f = a - b ;
        abs_error = fabs(f-s) ; 
        rel_error = abs_error/fabs(f) ;
        if ((rel_error > max_error) & (abs_error > max_error)){
            //sprintf(pt_serial_out_buffer, "a20=%e b20=%e \n\r", fp16_to_IEEE(a20), fp16_to_IEEE(b20));
            //serial_write ;
            sprintf(pt_serial_out_buffer, "diff16=%f error=%f rel.error=%f\n\r", s, abs_error, rel_error);
            serial_write ;
        }

        // mult test
        s = fp16_to_IEEE(fp16_mul(a16, b16)) ;
        f = a * b ;
        abs_error = fabs(f-s) ; 
        rel_error = abs_error/fabs(f) ;
        if ((rel_error > max_error) & (abs_error > max_error)){
            sprintf(pt_serial_out_buffer, "prod16=%f error=%f rel.error=%f\n\r", s, abs_error, rel_error);
            serial_write ;
        }
        
        // division test
        s = fp16_to_IEEE(fp16_div(a16, b16)) ;
        f = a / b ;
        abs_error = fabs(f-s) ; 
        rel_error = abs_error/fabs(f) ;
        if ((rel_error > max_error) & (abs_error > max_error)){
            sprintf(pt_serial_out_buffer, "ratio16=%f ieee=%f error=%f rel.error=%f\n\r", s, f, abs_error, rel_error);
            serial_write ;
        }

      
        // invsqrt test
        s = fp16_to_IEEE(fp16_invsqrt(a16)) ;
        f = 1/sqrt(a);
        abs_error = fabs(f-s) ; 
        rel_error = abs_error/fabs(f) ;
        if ((rel_error > max_error) & (abs_error > max_error)){
            sprintf(pt_serial_out_buffer, "invsqrt16=%f ieee=%f error=%f rel.error=%f\n\r", s, f, abs_error, rel_error);
            serial_write ;
        }
        
      }
      */
      //printf("%d\n\r", count) ;





      PT_YIELD_INTERVAL(100) ;
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // timer thread

// ==================================================
// === toggle25 thread 
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    
    // set up LED p25 to blink
    gpio_init(25) ;	
    gpio_set_dir(25, GPIO_OUT) ;
    gpio_put(25, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;
     
    while(1) {
      // yield time 0.1 second
      PT_YIELD_INTERVAL(200000) ;

      // toggle the LED on PICO
      LED_state = LED_state? false : true ;
      gpio_put(25, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === toggle_fast thread -- RUNNING on core 1
// ==================================================
// toggle gpio 4 
static PT_THREAD (protothread_toggle_fast(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    //
    // set up LED gpio 4 to blink
    gpio_init(4) ;	
    gpio_set_dir(4, GPIO_OUT) ;
    gpio_put(4, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;

    while(1) {
      //
      PT_YIELD_INTERVAL(20) ;
      // toggle gpio 4
      LED_state = !LED_state ;
      gpio_put(4, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_toggle_fast);
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");
     
  // === add threads ======================
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ===================
  // for core 0
  pt_add_thread(protothread_serial);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////