/*
 -- Multicore --
 floaitng point test code:


 */

#include "hardware/gpio.h"
#include "hardware/timer.h"

#include "pico/stdlib.h"
#include "hardware/uart.h"
#include "stdio.h"
#include <string.h>
#include <pico/multicore.h>
#include <pico/divider.h>
#include "hardware/sync.h"
#include "stdlib.h"
#include "math.h"

// =============================================
// fp20 format
// =============================================
// bit 19 is sign
// 6 bits 18-13 exponent
// 13 bits 12-0 mantissa (with one implied)
// represent one as 0|011111|0000000000000
// a zero exponent implies number=0
typedef unsigned int fp20 ;
//
fp20 IEEE_to_fp20(float a){
  fp20 b ;
  if(a==0) return 0 ;
  if(fabs(a)>5e9) a = 5e9 ;
  if(fabs(a)<1e-9) a = 1e-9 ;
  int ia = (*(int*)&a); ;
  int sgn = (ia>>31) ;
  int exponent = ((ia>>23) & 0xff) - 96 ; // -127 + 31
  int mantissa = ia & 0x7fffff ;
 // printf("%f %d %d %d\n\r", a, sgn, exponent, mantissa);
  b = (sgn<<19) | ((exponent)<<13)  | (mantissa>>10)  ; // >>23
  return b ;
}
// 
float fp20_to_IEEE(fp20 a){
  int ia = (int) a ;
  int sgn = (ia>>19) ;
  int exponent = (ia>>13) & 0x3f ;
  if (exponent==0) return 0.0 ;
  int mantissa = ia & 0x1fff ;
  float result = pow(2.0, (float) (exponent-31.0));
  result = (1.0+(((float)mantissa) / 8192.0)) * result;
  if(sgn) result = result * (-1);
  return result;
}

// =============================================
// negate and absolute
// =============================================
#define fp20_neg(a) ((a) ^ 0x080000)
#define fp20_abs(a) ((a) & 0x07ffff)

// =============================================
// add/subtract
// =============================================
fp20 fp20_add(fp20 a, fp20 b){
  // isolate mantissas and add one
  uint32_t aFrac = (a & 0x1fff) | 0x2000 ;
  uint32_t bFrac = (b & 0x1fff) | 0x2000 ;
  // isolate the exponents
  int aExp = (a >> 13) & 0x3f ;
  int bExp = (b >> 13) & 0x3f ;
  int expDiff = aExp - bExp ;
  // isolate the signs
  uint32_t aSign = a >> 19 ;
  uint32_t bSign = b >> 19 ;
  // add_sub is 1 for subtract
  uint32_t add_sub = aSign ^ bSign ;
  //
  if (aExp==0 | expDiff<-13) return b ;
  if (bExp==0 | expDiff>13) return a ;
  //
  if (expDiff>0){ // a bigger
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      // result is between 1 and 0.5
      uint32_t sum = aFrac - (bFrac>>expDiff) ;
      while (sum < 0x2000) {
        aExp--;
        sum <<= 1;
      }
      return (aSign<<19) | (aExp<<13) | (sum & 0x1fff)  ;
    }
    else { //add
      // result is between 1 and 2
      uint32_t sum = aFrac + (bFrac>>expDiff) ;
      if (sum >= 0x4000) {
        aExp++;
        sum >>= 1;
      }
      return (aSign<<19) | (aExp<<13) | (sum & 0x1fff) ;
    }
  }
  //
  if (expDiff<0){ // b bigger
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      uint32_t sum = bFrac - (aFrac>>-expDiff) ;
      while (sum < 0x2000) {
        bExp--;
        sum <<= 1;
      }
      return (bSign<<19) | (bExp<<13) | (sum & 0x1fff)  ;
    }
    else { //add
      uint32_t sum = bFrac + (aFrac>>-expDiff) ;
      if (sum >= 0x4000) {
        bExp++;
        sum >>= 1;
      }
      return (bSign<<19) | (bExp<<13) | (sum & 0x1fff) ;
    }
  }
  if (expDiff==0){ // 
    // add_sub is 1 for subtract
    if(add_sub){ // subtract
      // figure out which mantissa is bigger
      // compute bigger - smaller : result<1
      // normalize
      if (bFrac == aFrac) return 0 ;
      uint32_t sum = (bFrac > aFrac)? bFrac - aFrac : aFrac - bFrac ;
      while (sum < 0x2000) {
        bExp--;
        sum <<= 1;
        if (bExp == 0) return 0 ;
      }
      int Sign = (bFrac > aFrac)? bSign : aSign ;
      return (Sign<<19) | (bExp<<13) | (sum & 0x1fff)  ;
    }
    else { //add
      uint32_t sum = bFrac + aFrac ;
      // sum is between 2 and 4
      bExp++;
      sum >>= 1;
      return (bSign<<19) | (bExp<<13) | (sum & 0x1fff) ;
    }
  }
}

// ==================
fp20 fp20_sub(fp20 a, fp20 b){
  fp20 diff = fp20_add(a, fp20_neg(b)) ;
  return diff ;
}

// =============================================
// multiply
// =============================================
fp20 fp20_mul(fp20 a, fp20 b){
  // isolate mantissas and add one
  uint32_t aFrac = (a & 0x1fff) | 0x2000 ;
  uint32_t bFrac = (b & 0x1fff) | 0x2000 ;
  // isolate the exponents
  int aExp = (a >> 13) & 0x3f ;
  int bExp = (b >> 13) & 0x3f ;
  // add exponents an subtract offset
  int expSum = aExp + bExp - 31 ;
  // isolate the signs
  uint32_t aSign = a >> 19 ;
  uint32_t bSign = b >> 19 ;
  // get output sign
  uint32_t Sign = aSign ^ bSign ;
  //
  if (aExp==0 | bExp==0) return 0 ;
  if(expSum > 63) return 0 ;
  // range is 1 to 4
  uint32_t product = (aFrac * bFrac) >> 13 ;
  if (product >= 0x4000) {
    expSum++;
    product >>= 1;
    if(expSum > 63) return 0 ;
  }
  return (Sign<<19) | (expSum<<13) | (product & 0x1fff) ;
}

// =============================================
// divide
// =============================================
fp20 fp20_div(fp20 a, fp20 b){
  // isolate mantissas and add one
  uint32_t aFrac = (a & 0x1fff) | 0x2000 ;
  uint32_t bFrac = (b & 0x1fff) | 0x2000 ;
  // isolate the exponents
  int aExp = (a >> 13) & 0x3f ;
  int bExp = (b >> 13) & 0x3f ;
  // add exponents an subtract offset
  int expDiff = aExp - bExp + 31 ;
  // isolate the signs
  uint32_t aSign = a >> 19 ;
  uint32_t bSign = b >> 19 ;
  // get output sign
  uint32_t Sign = aSign ^ bSign ;
  //
  if (aExp==0 | bExp==0) return 0 ;
  if(expDiff< 1) return 0 ;
  // range is 2 to .5
  uint32_t ratio = (aFrac<<13) / bFrac  ;
  if (ratio < 0x2000) {
    expDiff--;
    ratio <<= 1;
    if(expDiff < 1) return 0 ;
  }
  return (Sign<<19) | (expDiff<<13) | (ratio & 0x1fff) ;
}

// ==========================================
// === protothreads globals
// ==========================================
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

// ==================================================
// === user's serial input thread
// ==================================================
// test speed of fixed point
static PT_THREAD (protothread_serial(struct pt *pt))
{
  PT_BEGIN(pt);
    static int num_iterations ;
    static int fptime, fp20time ;
    static float a, b ;
    static fp20 a20, b20 ;
    //
    while(1) {
      /*
      sprintf(pt_serial_out_buffer, "input 2 floats: ");
      serial_write ;
      serial_read ;
      // convert input string to number
      sscanf(pt_serial_in_buffer,"%f %f ", &a, &b) ;
      a20 = IEEE_to_fp20(a);
      b20 = IEEE_to_fp20(b);
      sprintf(pt_serial_out_buffer, "a20=%e b20=%e \n\r", fp20_to_IEEE(a20), fp20_to_IEEE(b20));
      serial_write ;

      // add test
      float s = fp20_to_IEEE(fp20_add(a20, b20)) ;
      float f = a + b;
      sprintf(pt_serial_out_buffer, "sum20=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
      serial_write ;

      // subtract test
      s = fp20_to_IEEE(fp20_sub(a20, b20)) ;
      f = a - b ;
      sprintf(pt_serial_out_buffer, "diff20=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
      serial_write ;

      // mult test
      s = fp20_to_IEEE(fp20_mul(a20, b20)) ;
      f = a * b ;
      sprintf(pt_serial_out_buffer, "prod20=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
      serial_write ;
      
      // division test
      s = fp20_to_IEEE(fp20_div(a20, b20)) ;
      f = a / b ;
      sprintf(pt_serial_out_buffer, "ratio20=%f error=%f rel.error=%f\n\r", s, fabs(f-s), fabs(f-s)/fabs(f));
      serial_write ;
     */

      // print prompt
      sprintf(pt_serial_out_buffer, "input iterations ");
      // spawn a thread to do the non-blocking write
      serial_write ;

      // spawn a thread to do the non-blocking serial read
      serial_read ;
      // convert input string to number
      sscanf(pt_serial_in_buffer,"%d ", &num_iterations) ;

      //do fp add
      float fp1 = 0;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = 10.0 + fp1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do fp20 add 
      fp20 d20 = IEEE_to_fp20(0.0);
      fp20 ten = IEEE_to_fp20(10.0);
      fp20time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        d20 = fp20_add(ten, (d20)) ;
      }
      fp20time = PT_GET_TIME_usec() - fp20time ;

      printf("fptime %d    ieee sum %f \r\n", fptime, fp1) ;
      printf("fp20time %d ffp20 sum %f \r\n", fp20time, fp20_to_IEEE(d20));
      printf("\r\n") ;

      //do fp mult
      fp1 = 1.0 ;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = 1.01 * fp1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do fp20 mult 
      d20 = IEEE_to_fp20(1.0);
      fp20 pc = IEEE_to_fp20(1.01);
      fp20time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        d20 = fp20_mul(pc, (d20)) ;
      }
      fp20time = PT_GET_TIME_usec() - fp20time ;

      printf("fptime %d    ieee prod %f \r\n", fptime, fp1) ;
      printf("fp20time %d ffp20 prod %f \r\n", fp20time, fp20_to_IEEE(d20));
      printf("\r\n") ;

      //do fp div
      fp1 = 1.0 ;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = 10.0 / fp1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do fp20 div 
      d20 = IEEE_to_fp20(1.0);
      pc = IEEE_to_fp20(10.);
      fp20time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        d20 = fp20_div(pc, d20) ;
      }
      fp20time = PT_GET_TIME_usec() - fp20time ;

      printf("fptime %d    ieee ratio %f \r\n", fptime, fp1) ;
      printf("fp20time %d ffp20 ratio %f \r\n", fp20time, fp20_to_IEEE(d20));
      printf("\r\n") ;

/*
      //do s1x14 MAC
      s1x14 e1 = float_to_s1x14(0.01), fix14_1 = float_to_s1x14(.5);
      fix14time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix14_1 = fix14_1 + muls1x14(fix14_1, e1) ;
      }
      fix14time = PT_GET_TIME_usec() - fix14time ;
      
      // timeing varies, but generally s15x16 is 3 to 4 times
      // faster than floaqt and s1x14 is 10 to 12 times faster
      // print summary of timing
      printf("fptime %d  fp_mac %f \r\n", fptime, fp1) ;
      printf("fix16time %d fix_mac %f \r\n", fixtime, s15x16_to_float(fix1));
      printf("fix14time %d fix_mac %f \r\n", fix14time, s1x14_to_float(fix14_1));
      printf("\r\n") ;
      // =================
      //do fp divide
      c1 = 1.01 ; 
      fp1 = 1.0;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = fp1 / c1 ;
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do s15x16 divide
      d1 = float_to_s15x16(1.01);
       fix1 = float_to_s15x16(1.0);
      fixtime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix1 = divs15x16(fix1, d1) ;
      }
      fixtime = PT_GET_TIME_usec() - fixtime ;

      //do s1x14 
      e1 = float_to_s1x14(1.01);
      fix14_1 = float_to_s1x14(1.0);
      fix14time = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix14_1  =  divs1x14(fix14_1, e1) ;
      }
      fix14time = PT_GET_TIME_usec() - fix14time ;
      
      // timeing varies, but generally s15x16 is 3 to 4 times
      // faster than floaqt and s1x14 is 10 to 12 times faster
      // print summary of timing
      printf("fptime %d  fp_div %f \r\n", fptime, fp1) ;
      printf("fix16time %d fix_div %f \r\n", fixtime, s15x16_to_float(fix1));
      printf("fix14time %d fix_div %f \r\n", fix14time, s1x14_to_float(fix14_1));
      printf("\r\n") ;

// =================
      //do fp sqrt
      c1 = 1.01 ; 
      fp1 = 1.0;
      fptime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fp1 = fp1 + sqrt(fp1);
      }
      fptime = PT_GET_TIME_usec() - fptime ;

      //do s15x16 sqrt
      d1 = float_to_s15x16(1.01);
      fix1 = float_to_s15x16(1.0);
      fixtime = PT_GET_TIME_usec();
      for(int i=0; i<num_iterations; i++){
        fix1 = fix1 + sqrts15x16(fix1) ;
      }
      fixtime = PT_GET_TIME_usec() - fixtime ;

      // timeing varies
      printf("fptime %d  fp_sqrt %f \r\n", fptime, fp1) ;
      printf("fix16time %d fix_sqrt %f \r\n", fixtime, s15x16_to_float(fix1));
      //printf("fix14time %d fix_div %f \r\n", fix14time, s1x14_to_float(fix14_1));
      printf("\r\n") ;
*/

      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // timer thread

// ==================================================
// === toggle25 thread 
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    
    // set up LED p25 to blink
    gpio_init(25) ;	
    gpio_set_dir(25, GPIO_OUT) ;
    gpio_put(25, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;
     
    while(1) {
      // yield time 0.1 second
      PT_YIELD_INTERVAL(100000) ;

      // toggle the LED on PICO
      LED_state = LED_state? false : true ;
      gpio_put(25, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === toggle_fast thread -- RUNNING on core 1
// ==================================================
// toggle gpio 4 
static PT_THREAD (protothread_toggle_fast(struct pt *pt))
{
  PT_BEGIN(pt);
    static bool LED_state = false ;
    //
    // set up LED gpio 4 to blink
    gpio_init(4) ;	
    gpio_set_dir(4, GPIO_OUT) ;
    gpio_put(4, true);
    // data structure for interval timer
    PT_INTERVAL_INIT() ;

    while(1) {
      //
      PT_YIELD_INTERVAL(20) ;
      // toggle gpio 4
      LED_state = !LED_state ;
      gpio_put(4, LED_state);
      //
      // NEVER exit while
    } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_toggle_fast);
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");
     
  // === add threads ======================
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ===================
  // for core 0
  pt_add_thread(protothread_serial);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main
///////////
// end ////
///////////